"""
MSeed object to hold and process MSeed data
"""
import os
from pathlib import Path
from typing import Dict, Tuple, List, Set

from obspy.core import Stream

from sohstationviewer.controller.util import validate_file, validate_dir
from sohstationviewer.model.data_type_model import DataTypeModel, ThreadStopped
from sohstationviewer.model.handling_data import (
    squash_gaps, sort_data, read_mseed_trace_spr_less_than_or_equal_1,
    read_mseed_or_text)

from sohstationviewer.view.util.enums import LogType


class MSeed(DataTypeModel):
    """
    read and process mseed file into object with properties can be used to
    plot SOH data, mass position data, waveform data and gaps
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # channels: available channels got from files among
        # requested channels
        self.channels: Set[str] = set()

        # sta_ids: available sta_ids got from files
        self.sta_ids: Set[str] = set()

        """
        A station can have more than one netcode which should be consider
        a wrong setting. This will be warned to user, but just implementation
        won't have any special work on that.
        """
        self.nets_by_sta: Dict[str, List[str]] = {}

        if self.creator_thread.isInterruptionRequested():
            raise ThreadStopped()
        self.get_soh_masspos_waveform_gaps(self.dir)

        if self.creator_thread.isInterruptionRequested():
            raise ThreadStopped()
        self.selected_key = self.select_sta_id()
        if self.selected_key is None:
            raise ThreadStopped()
        if (len(self.req_wf_chans) != 0 and
                self.selected_key in self.waveform_data.keys()):
            sort_data(self.waveform_data[self.selected_key]['readData'])

        self.check_not_found_soh_chans()

    def get_soh_masspos_waveform_gaps(self, folder: str):
        """
            + read waveform data for filename associate with time range
            + read soh_stream for unmerged traces
            + merge soh_stream into SOH data and mass position data
            + squash gaps from different soh channels to one
        :param folder: absolute path to data set folder
        """

        (self.waveform_data,
         soh_streams,
         mass_pos_streams) = self.read_folder(folder)

        gaps = {}
        self.soh_data = self.merge_streams(
            soh_streams, gaps, is_mass_pos=False)
        self.mass_pos_data = \
            self.merge_streams(mass_pos_streams, gaps, is_mass_pos=True)
        for sta_id in gaps.keys():
            self.gaps[sta_id] = squash_gaps(gaps[sta_id])

    def read_folder(self, folder: str
                    ) -> Tuple[Dict, Dict[str, Dict[str, Stream]]]:
        """
        Read data streams for soh, mass position and waveform.
        For waveform, traces in data streams will be turn into memmap file
            to save resources
        :param folder: absolute path to data set folder
        :return waveform_data: a dict for waveform data including
        {sta_id: {'readData': { chan_id: {samplerate
                         and preset traces_info which is a list of {info and
                         np.memmap's file path of trace data and time}
                  }
        }
            (use np.memmap because data are too large to be merged and
                processed)
        :return soh_streams: a dict of soh channel streams
            {sta_id: {chan_id: mseed stream of traces}}
        """
        soh_streams = {}
        mass_pos_streams = {}
        waveform_data = {}
        stat_prop = {}
        chan_prop = {}
        count = 0
        total = sum([len(files) for _, _, files in os.walk(folder)])
        for path, sub_dirs, files in os.walk(folder):
            try:
                validate_dir(path)
            except Exception as e:
                # skip Information folder
                self.track_info(str(e), LogType.WARNING)
                continue

            for file_name in files:
                if self.creator_thread.isInterruptionRequested():
                    raise ThreadStopped()

                path2file = Path(path).joinpath(file_name)
                if not validate_file(path2file, file_name):
                    continue
                count += 1
                if count % 10 == 0:
                    self.track_info(
                         f'Read {count} files/{total}', LogType.INFO)
                try:
                    ret = read_mseed_or_text(
                        path2file, self.tmp_dir, self.read_start,
                        self.read_end,
                        soh_streams, mass_pos_streams, waveform_data,
                        self.data_time, self.log_data,
                        self.req_soh_chans, self.req_wf_chans,
                        self.include_mp123zne, self.include_mp456uvw,
                        self.track_info)
                except Exception as e:
                    self.track_info(str(e), LogType.WARNING)
                    continue

                if ret is None:
                    continue

                self.sta_ids.update(ret['sta_ids'])

                if len(ret['sta_ids']) > 1:
                    k = tuple(ret['sta_ids'])
                    if k not in stat_prop:
                        stat_prop[k] = []
                    stat_prop[k].append(file_name)

                if len(ret['chan_ids']) > 1:
                    k = tuple(ret['chan_ids'])
                    if k not in chan_prop:
                        chan_prop[k] = []
                    chan_prop[k].append(file_name)
                self.channels.update(ret['chan_ids'])
                if 'read' not in ret:
                    # not waveform
                    continue
        self.track_info(
            f'Skipped {total - count} invalid files.', LogType.INFO)
        if len(stat_prop) > 0:
            errmsg = (f"More than one stations in a file: {stat_prop}. "
                      f"Will use the first one.")
            self.track_info(errmsg, LogType.ERROR)

        return waveform_data, soh_streams, mass_pos_streams

    def merge_streams(self, streams: Dict[str, Dict[str, Stream]],
                      gaps: Dict, is_mass_pos) -> Dict:
        """
        Merge traces in the given steams
        :param streams: a dict of soh channel streams
            {sta_id: {chan_id: mseed stream of traces}}
        if there is more than one network for a station, user can either choose
        data for one network for that station, or combine all networks related
        to that station into one data set (by changing network name to the
        selected network then merge again)

        :return data_dict: dict of trace info by chan_id and sta_id
        """

        data_dict = {}
        all_gaps: List[List[float]] = []
        sta_ids = []
        for sta_id in streams:
            self.sta_ids.add(sta_id)
            data_dict[sta_id] = {}
            self.nets_by_sta[sta_id] = set()
            for chan_id in streams[sta_id]:
                self.channels.add(chan_id)
                stream = streams[sta_id][chan_id]
                nets = set(tr.stats['network'].strip() for tr in stream)
                if len(nets) > 1:
                    for tr in stream:
                        tr.stats['network'] = stream[0].stats['network']
                self.nets_by_sta[sta_id].update(nets)

                stream.merge()
                tr = stream[0]

                gaps_in_stream = stream.get_gaps()
                all_gaps += [[g[4].timestamp, g[5].timestamp]
                             for g in gaps_in_stream]
                trace_info = read_mseed_trace_spr_less_than_or_equal_1(tr)

                if is_mass_pos:
                    data_dict[sta_id][chan_id] = {
                        'orgTrace': trace_info}
                else:
                    data_dict[sta_id][chan_id] = {
                        'orgTrace': trace_info,
                        'samplerate': trace_info['samplerate']}
            if sta_id not in gaps:
                gaps[sta_id] = []
                self.gaps[sta_id] = []
                sta_ids.append(sta_id)
            gaps[sta_id] += all_gaps
        return data_dict

    def select_sta_id(self) -> str:
        """
        :return selected_sta_id: the selected station id from self.sta_ids
            + If there is only one station id, return it.
            + If there is more than one, show all ids, let user choose one to
                return.
        """

        sta_ids = list(self.sta_ids)
        selected_sta_id = sta_ids[0]
        if len(sta_ids) > 1:
            msg = ("There are more than one stations in the given data.\n"
                   "Please select one to display")
            self.pause_signal.emit(msg, sta_ids)
            self.pause()
            selected_sta_id = sta_ids[self.pause_response]

        self.track_info(f'Select Station {selected_sta_id}', LogType.INFO)
        return selected_sta_id
